import hashlib

def generate_short_md5(input_string, len = 8):
    md5_hash = hashlib.md5()
    md5_hash.update(input_string.encode('utf-8'))
    return md5_hash.hexdigest()[:len]

def is_a_share(stock_code):
    """
    判断给定的股票代码是否是A股代码且不含字母。

    Args:
    stock_code (str): 股票代码

    Returns:
    bool: 如果是A股代码且不含字母则返回True，否则返回False
    """
    # 检查股票代码长度是否为6位，且全部由数字组成
    if len(stock_code) == 6 and stock_code.isdigit():
        # 检查股票代码是否以0, 3, 6开头
        if stock_code.startswith(("0", "3", "6")):
            return True
    return False


from .stock_base_loader import StockBaseLoader
import numpy as np
from io import BytesIO
import matplotlib.pyplot as plt
import mplfinance as mpf
import pandas as pd
from PIL import Image


class StockKlineLoader(StockBaseLoader):
    def __init__(self, symbol, show_chart = False, figsize=(3, 3), categtory_nums = None, **kwargs):
        self.show_chart = show_chart
        self.figsize = figsize
        if categtory_nums is None:
            categtory_nums = (5, 2, -2, -5)
        self.intervals = categtory_nums
        super().__init__(symbol, **kwargs)
        
    def set_feature_cols(self):
        self.feature_cols = self.price_cols + self.other_cols + ['日期']
        return self
        
    def cal_features_labels(self, i, row_data, data):
        feature_start = i
        feature_end = i + self.sequence_length - 1
        label_start = feature_end
        label_end = label_start + self.predict_length - 1

        x_df = data.loc[feature_start:feature_end, self.feature_cols].copy()
        y_df = data.loc[label_start:label_end, self.label_cols].copy()

        max_price = x_df['最高'].max()
        min_price = x_df['最低'].min()

        # Normalize prices
        x_df.loc[:, self.price_cols] = x_df[self.price_cols].apply(
            lambda x: self.normalize_price(max_price, min_price, x)
        )
        # Normalize other features
        x_df.loc[:, self.other_cols] = x_df[self.other_cols].apply(self.min_max_normalize)
        y_df.loc[:, self.label_cols] = y_df[self.label_cols].apply(
            lambda x: self.normalize_price(max_price, min_price, x)
        )

        # 计算涨幅百分比
        pre_price = x_df.iloc[-1]['收盘']
        # 检查 pre_price 是否为零
        if pre_price == 0:
            price_change_percentage = y_df['收盘'].mean() * 100
        else:
            price_change_percentage = ((y_df['收盘'].mean() - pre_price) / pre_price) * 100

        intervals = self.intervals
        result = np.zeros(5)  # 创建一个零数组来存类别概率
        # 使用任何方法对于Series的比较
        if price_change_percentage >= intervals[0]:
            result[0] = 1
        elif price_change_percentage >= intervals[1]:
            result[1] = 1
        elif price_change_percentage >= intervals[2]:
            result[2] = 1
        elif price_change_percentage >= intervals[3]:
            result[3] = 1
        else:
            result[4] = 1

        img_array = self.generate_kline_image(x_df)
        return img_array, result


    
    def generate_kline_image(self, df):
        # 确保索引为 DatetimeIndex
        df.index = pd.to_datetime(df['日期'])
        df.rename(columns={
            '开盘': 'Open',
            '收盘': 'Close',
            '最高': 'High',
            '最低': 'Low',
            '成交量': 'Volume'
        }, inplace=True)


        mc = mpf.make_marketcolors(up='red', down='green', inherit=True)
        s = mpf.make_mpf_style(marketcolors=mc, base_mpf_style='yahoo')
        
        fig, ax = plt.subplots(figsize=self.figsize)
        
        try:
            mpf.plot(df, type='candle', ax=ax, show_nontrading=False, style=s)
        except Exception as e:
            print("Error plotting the chart:", e)  # 捕获并打印错误信息

        buf = BytesIO()
        plt.axis('off')
        plt.savefig(buf, format='png', bbox_inches='tight', pad_inches=0.5)
        buf.seek(0)
        
        img = Image.open(buf)
        img_array = np.array(img)
        if self.show_chart:
            plt.show()
        plt.close(fig)
        return img_array / 255.0

    



from torch.utils.data import Dataset

class StockDataset(Dataset):
    def __init__(self, X, Y, transform=None):
        self.X = X
        self.Y = Y
        self.transform = transform

    def __getitem__(self, index):
        x = self.X[index]
        y = self.Y[index]
        if self.transform:
            x = self.transform(x)
        return x, y

    def __len__(self):
        return len(self.X)


from china_stock_data import StockData
from .stock_base_loader import StockBaseLoader
import math
class StockTrendLoader(StockBaseLoader):
    def __init__(self, symbol, **kwargs):
        super().__init__(symbol, **kwargs)
        
    def set_feature_cols(self):
        self.feature_cols = self.price_cols + self.other_cols
        return self

    def cal_features_labels(self, i, row_data, data):
        feature_start = i
        feature_end = i + self.sequence_length - 1
        label_start = feature_end
        label_end = label_start + self.predict_length - 1
        x_df = data.loc[feature_start:feature_end, self.feature_cols].copy()
        y_df = data.loc[label_start:label_end, self.label_cols].copy()
        
        max_price = math.ceil(x_df['最高'].max())
        min_price = math.floor(x_df['最低'].min())

        # Normalize prices
        x_df.loc[:, self.price_cols] = x_df[self.price_cols].apply(
            lambda x: self.normalize_price(max_price, min_price, x)
        )
        # Normalize other features
        x_df.loc[:, self.other_cols] = x_df[self.other_cols].apply(self.min_max_normalize)
        # Normalize labels
        y_df.loc[:, self.label_cols] = y_df[self.label_cols].apply(
            lambda x: self.normalize_price(max_price, min_price, x)
        )
        y_df = y_df.mean()
        return x_df, y_df
    
    



from sklearn.model_selection import train_test_split
import numpy as np
from .stock_dataset import StockDataset
from torch.utils.data import DataLoader

class BaseLoader:
    def __init__(self, sequence_length=30, predict_length=1, batch_size=32, test_ratio=0.1):
        self.predict_length = predict_length
        self.sequence_length = sequence_length
        self.batch_size = batch_size
        self.test_ratio = test_ratio
        
        
    def cal_features_labels(self, i, row_data, data):
        """
        计算特征值

        :param i: 当前行索引
        :param row_data: 当前行数据
        :param data: 整个数据集
        :return: 特征值
        """
        raise NotImplementedError("Subclasses must implement the cal_features method.")
    
    def create_sequences(self):
        if not hasattr(self, 'data'):
            self.format_data()
        X, Y = [], []
        length = len(self.data)
        for i in range(length - self.sequence_length - self.predict_length):
            x, y = self.cal_features_labels(i, self.data.iloc[i], self.data)
            X.append(x)
            Y.append(y)

        arr_x = np.array(X, dtype=np.float32)
        arr_y = np.array(Y, dtype=np.float32)
        return (arr_x, arr_y)

    def get_dataset(self):
        if not hasattr(self, 'train_dataset') or not hasattr(self, 'test_dataset'):
            X, Y = self.create_sequences()
            X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=self.test_ratio, shuffle=False)
            self.train_dataset = StockDataset(X_train, Y_train)
            self.test_dataset = StockDataset(X_test, Y_test)
        return self.train_dataset, self.test_dataset

    def get_train_loader(self):
        train_dataset,_ = self.get_dataset()
        return DataLoader(dataset=train_dataset, batch_size=self.batch_size, shuffle=False, num_workers = 4)

    def get_test_loader(self):
        _ ,test_dataset = self.get_dataset()
        return DataLoader(dataset=test_dataset, batch_size=self.batch_size, shuffle=False, num_workers = 4)
    
    def get_recent_data(self):
        if not hasattr(self, 'data'):
            self.format_data()
        idx = len(self.data) - self.sequence_length
        x_df, _ = self.cal_features_labels(idx, self.data.iloc[idx], self.data)
        return np.array(x_df, dtype=np.float32)
        
    



from .stock_trend_loader import StockTrendLoader
from china_stock_data import StockMarket
from .utils import is_a_share
class IndexTrendLoader:
    def __init__(self, symbols = None, index = None, start = None, limit = None, **kwargs):
        self.stocks: dict[str, StockTrendLoader] = {}
        self.kwargs = kwargs
        if symbols is not None:
            self.symbols = symbols
            for symbol in symbols:
                self.stocks[symbol] = StockTrendLoader(symbol, **kwargs)
        if index:
            self.stock_market = StockMarket(index)
            codes = self.stock_market['index_codes']

            if start is None:
                start = 0
            if limit is None:
                limit = len(codes)

            actual_limit = min(start + limit, len(codes))

            for i in range(start, actual_limit):
                symbol = codes[i]
                if symbol not in self.stocks and is_a_share(symbol):
                    self.stocks[symbol] = StockTrendLoader(symbol=symbol, **kwargs)
                    



from china_stock_data import StockData
from .base_loader import BaseLoader
import math
class StockBaseLoader(BaseLoader):
    def __init__(self, symbol, **kwargs):
        self.symbol = symbol
        self.stock = StockData(symbol, days=365 * 10)
        self.price_cols = ['开盘', '收盘', '最高', '最低']
        self.other_cols = ['成交量']
        self.label_cols = ['收盘']
        self.set_feature_cols()
        super().__init__(**kwargs)
        
    def origin_data(self):
        """
        从china_stock_data加载原始的数据，这里只加载了历史交易数据，K线数据
        :return: self对象
        """
        self.origin = self.stock.kline
        return self
    
    def format_data(self):
        """
        格式化数据，包括加载原始数据、处理数据、计算最高和最低价格，并将数据存储在self.data中
        :return: self对象
        """
        if not hasattr(self, 'origin'):
            self.origin_data()
        data = self.origin.copy()
        data.drop(['股票代码'], axis=1, inplace=True)
        self.max_price = math.ceil(data['最高'].max())
        self.min_price = math.floor(data['最低'].min())
        data['成交量'] = data['成交量'].astype(float)
        self.data = data
        return self
    
    def normalize_price(self, max_price, min_price, val):
        return (val - min_price) / (max_price - min_price)
    
    def unnormalize_price(self, val):
        return val * (self.max_price - self.min_price) + self.min_price
    
    def min_max_normalize(self, series):
        return (series - series.min()) / (series.max() - series.min())
    
    def cal_features_labels(self, i, row_data, data):
        feature_start = i
        feature_end = i + self.sequence_length - 1
        label_start = feature_end
        label_end = label_start + self.predict_length - 1
        x_df = data.loc[feature_start:feature_end, self.feature_cols].copy()
        y_df = data.loc[label_start:label_end, self.label_cols].copy()
        
        max_price = math.ceil(x_df['最高'].max())
        min_price = math.floor(x_df['最低'].min())

        # Normalize prices
        x_df.loc[:, self.price_cols] = x_df[self.price_cols].apply(
            lambda x: self.normalize_price(max_price, min_price, x)
        )
        # Normalize other features
        x_df.loc[:, self.other_cols] = x_df[self.other_cols].apply(self.min_max_normalize)
        # Normalize labels
        y_df.loc[:, self.label_cols] = y_df[self.label_cols].apply(
            lambda x: self.normalize_price(max_price, min_price, x)
        )
        y_df = y_df.mean()
        return x_df, y_df

    
    def load_data(self):
        """
        返回类似这样的数据
        :return: (x_train, y_train), (x_test, y_test)
        """
        if not hasattr(self, 'data'):
            self.format_data()
        
            
    def load_recent(self):
        """
        返回类似这样的数据
        :return: (x_train, y_train), (x_test, y_test)
        """
        if not hasattr(self, 'data'):
            self.format_data()     
            
    
    



from .base_loader import BaseLoader
from .stock_trend_loader import StockTrendLoader
from .stock_base_loader import StockBaseLoader
from .stock_kline_loader import StockKlineLoader
from .index_trend_loader import IndexTrendLoader
from .utils import is_a_share